/************************************************************************
 *
 *  Application.java
 *
 *  Copyright: 2002-2015 by Henrik Just
 *
 *  This file is part of Writer2LaTeX.
 *  
 *  Writer2LaTeX is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  Writer2LaTeX is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with Writer2LaTeX.  If not, see <http://www.gnu.org/licenses/>.
 * 
 *  Version 1.6 (2015-01-09) 
 *
 */

package w2phtml;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileNotFoundException;
//import java.io.FileOutputStream;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Set;
import java.util.Vector;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import pro.litvinovg.w2phtml.gui.ConfigurationWindow;
import w2phtml.api.Converter;
import w2phtml.api.ConverterFactory;
import w2phtml.api.ConverterResult;
import w2phtml.api.MIMETypes;
import w2phtml.util.Misc;

public final class Application {

	private static final Logger logger = LoggerFactory.getLogger(Application.class);
	/* Based on command-line parameters. */
	private String sTargetMIME = MIMETypes.HTML5;
	private Vector<String> configFileNames = new Vector<String>();
	private String sTemplateFileName = null;
	private String sStyleSheetFileName = null;
	private Set<String> resources = new HashSet<String>();
	private Hashtable<String, String> options = new Hashtable<String, String>();
	private String sSource = null;
	private String sTarget = null;

	/**
	 * Main method
	 *
	 * @param args
	 *          The argument passed on the command line.
	 */
	public static final void main(String[] args) throws Exception {
		if (args.length < 2) {
				showUsage("");
				String fileName = "";
				if (args.length == 1){
					fileName = args[0];
				}
				ConfigurationWindow.runGUI(fileName);	
		} else {
			try {
				long time = System.currentTimeMillis();
				Application app = new Application();
				app.parseCommandLine(args);
				app.doConversion();
				logger.debug("Total conversion time was " + (System.currentTimeMillis() - time) + " miliseconds");
			} catch (IllegalArgumentException ex) {
				String msg = ex.getMessage();
				showUsage(msg);
			}
		}

	}

	// Convert the directory or file
	private void doConversion() throws Exception {
		sayHello();
		File source = new File(sSource);
		examineSource(source);

		// Step 3: Examine target
		File target;
		target = examineTarget(source);

		// Step 4: Create converters
		Converter converter = ConverterFactory.createConverter(sTargetMIME);
		if (converter == null) {
			logger.error("Failed to create converter for " + sTargetMIME);
			throw new Exception("Conversion failed.");
		}

		// Step 5a: Read template
		if (sTemplateFileName != null) {
			try {
				logger.info("Reading template " + sTemplateFileName);
				byte[] templateBytes = Misc.inputStreamToByteArray(new FileInputStream(sTemplateFileName));
				converter.readTemplate(new ByteArrayInputStream(templateBytes));
			} catch (FileNotFoundException e) {
				logger.error("--> This file does not exist!");
				logger.error("    " + e.getMessage());
			} catch (IOException e) {
				logger.error("--> Failed to read the template file!");
				logger.error("    " + e.getMessage());
			}
		}

		// Step 5b: Read style sheet
		if (sStyleSheetFileName != null) {
			try {
				logger.info("Reading style sheet " + sStyleSheetFileName);
				byte[] styleSheetBytes = Misc.inputStreamToByteArray(new FileInputStream(sStyleSheetFileName));
				converter.readStyleSheet(new ByteArrayInputStream(styleSheetBytes));
			} catch (FileNotFoundException e) {
				logger.error("--> This file does not exist!");
				logger.error("    " + e.getMessage());
			} catch (IOException e) {
				logger.error("--> Failed to read the style sheet file!");
				logger.error("    " + e.getMessage());
			}
		}

		// Step 5c: Read style resources
		for (String sResource : resources) {
			String sMediaType;
			String sFileName;
			int nSeparator = sResource.indexOf("::");
			if (nSeparator > -1) {
				sFileName = sResource.substring(0, nSeparator);
				sMediaType = sResource.substring(nSeparator + 2);
			} else {
				sFileName = sResource;
				sMediaType = null;
			}
			logger.info("Reading resource file " + sFileName);
			try {
				byte[] resourceBytes = Misc.inputStreamToByteArray(new FileInputStream(sFileName));
				converter.readResource(new ByteArrayInputStream(resourceBytes), sFileName, sMediaType);
			} catch (IOException e) {
				logger.error("--> Failed to read the resource file!");
				logger.error("    " + e.getMessage());
			}

		}

		// Step 6: Read config
		for (int i = 0; i < configFileNames.size(); i++) {
			String sConfigFileName = (String) configFileNames.get(i);
			if (sConfigFileName.startsWith("*")) {
				sConfigFileName = sConfigFileName.substring(1);
				logger.info("Reading default configuration " + sConfigFileName);
				try {
					converter.getConfig().readDefaultConfig(sConfigFileName);
				} catch (IllegalArgumentException e) {
					logger.error("--> This configuration is unknown!");
					logger.error("    " + e.getMessage());
				}
			} else {
				logger.info("Reading configuration file " + sConfigFileName);
				try {
					byte[] configBytes = Misc.inputStreamToByteArray(new FileInputStream(sConfigFileName));
					converter.getConfig().read(new ByteArrayInputStream(configBytes));

				} catch (IOException e) {
					logger.error("--> Failed to read the configuration!");
					logger.error("    " + e.getMessage());
				}
			}
		}

		// Step 7: Set options from command line
		Enumeration<String> keys = options.keys();
		while (keys.hasMoreElements()) {
			String sKey = keys.nextElement();
			String sValue = (String) options.get(sKey);
			converter.getConfig().setOption(sKey, sValue);
		}

		// Step 8: Perform conversion

		logger.info("Converting " + source.getPath());
		ConverterResult dataOut = null;

		try {
			dataOut = converter.convert(source, target.getName());
		} catch (FileNotFoundException e) {
			logger.error("--> The file " + source.getPath() + " does not exist!");
			logger.error("    " + e.getMessage());
			throw new Exception("Conversion failed.");
		} catch (IOException e) {
			logger.error("--> Failed to convert the file " + source.getPath() + "!");
			logger.error("    " + e.getMessage());
			logger.error("    Please make sure the file is in OpenDocument format");
			throw new Exception("Conversion failed.");
		}

		// TODO: Should do some further checking on the feasability of writing
		// the directory and the files.
		File targetDir = target.getParentFile();
		if (targetDir != null && !targetDir.exists()) {
			targetDir.mkdirs();
		}
		try {
			dataOut.write(targetDir);
		} catch (IOException e) {
			logger.error("--> Error writing out file!");
			logger.error("    " + e.getMessage());
			throw new Exception("Conversion failed.");
		}

		// Step 9: Say goodbye!
		logger.info("Done!");
	}

	private File examineTarget(File source) {
		File target;

		if (sTarget == null) {
			target = new File(source.getParent(), Misc.removeExtension(source.getName()));
		} else {
			target = new File(sTarget);
			if (sTarget.endsWith(File.separator)) {
				target = new File(target, Misc.removeExtension(source.getName()));
			}
		}

		return target;
	}

	private void examineSource(File source) throws Exception {
		if (!source.exists()) {
			logger.error("I'm sorry, I can't find " + sSource);
			throw new Exception("Conversion failed.");
		}
		if (!source.canRead()) {
			logger.error("I'm sorry, I can't read " + sSource);
			throw new Exception("Conversion failed.");
		}
	}

	private void sayHello() {
		logger.info("This is w2phtml, Version " + ConverterFactory.getVersion() + " (" + ConverterFactory.getDate() + ")");
		logger.info("Starting conversion...");
	}

	/**
	 * Display usage.
	 */
	private static void showUsage(String msg) {
		System.out.println();
		System.out.println("This is w2phtml, Version " + ConverterFactory.getVersion() + " (" + ConverterFactory.getDate() + ")");
		System.out.println();
		if (msg != null)
			System.out.println(msg);
		System.out.println();
		System.out.println("Usage:");
		System.out.println("   java -jar <path>/w2phtml.jar <options> <source file/directory> [<target file/directory>]");
		System.out.println("where the available options are:");
		System.out.println("   -xhtml");
		System.out.println("   -xhtml11");
		System.out.println("   -xhtml+mathml");
		System.out.println("   -html5");
		System.out.println("   -epub");
		System.out.println("   -epub3");
		System.out.println("   -template[=]<template file>");
		System.out.println("   -stylesheet[=]<style sheet file>");
		System.out.println("   -resource[=]<resource file>[::<media type>]");
		System.out.println("   -cleanxhtml");
		System.out.println("   -config[=]<configuration file>");
		System.out.println("   -<configuration option>[=]<value>");
		System.out.println("See the documentation for the available configuration options");
	}

	/**
	 * Parse command-line arguments.
	 *
	 * @param args
	 *          Array of command line arguments.
	 *
	 * @throws IllegalArgumentException
	 *           If an argument is invalid.
	 */
	private void parseCommandLine(String sArgs[]) throws IllegalArgumentException {

		int i = 0;

		while (i < sArgs.length) {
			String sArg = getArg(i++, sArgs);
			if (sArg.startsWith("-")) { // found an option
				if ("-html5".equals(sArg)) {
					sTargetMIME = MIMETypes.HTML5;
				} else if ("-xhtml".equals(sArg)) {
					sTargetMIME = MIMETypes.XHTML;
				} else if ("-xhtml11".equals(sArg)) {
					sTargetMIME = MIMETypes.XHTML11;
				} else if ("-xhtml+mathml".equals(sArg)) {
					sTargetMIME = MIMETypes.XHTML_MATHML;
				} else if ("-epub".equals(sArg)) {
					sTargetMIME = MIMETypes.EPUB;
				} else if ("-epub3".equals(sArg)) {
					sTargetMIME = MIMETypes.EPUB3;
				} else if ("-rdf".equals(sArg)) {
					sTargetMIME = MIMETypes.RDF;
				} else if ("-cleanxhtml".equals(sArg)) {
					configFileNames.add("*cleanxhtml.xml");
				} else { // option with argument
					int j = sArg.indexOf("=");
					String sArg2;
					if (j > -1) { // argument is separated by =
						sArg2 = sArg.substring(j + 1);
						sArg = sArg.substring(0, j);
					} else { // argument is separated by space
						sArg2 = getArg(i++, sArgs);
					}
					if ("-config".equals(sArg)) {
						configFileNames.add(sArg2);
					} else if ("-template".equals(sArg)) {
						sTemplateFileName = sArg2;
					} else if ("-stylesheet".equals(sArg)) {
						sStyleSheetFileName = sArg2;
					} else if ("-resource".equals(sArg)) {
						resources.add(sArg2);
					} else { // configuration option
						options.put(sArg.substring(1), sArg2);
					}
				}
			} else { // not an option, so this must be the source
				sSource = sArg;
				// Possibly followed by the target
				if (i < sArgs.length) {
					String sArgument = getArg(i++, sArgs);
					if (sArgument.length() > 0) {
						sTarget = sArgument;
					}
				}
				// Skip any trailing empty arguments and signal an error if there's more
				while (i < sArgs.length) {
					String sArgument = getArg(i++, sArgs);
					if (sArgument.length() > 0) {
						throw new IllegalArgumentException("I didn't expect " + sArgument + "?");
					}
				}
			}
		}
		if (sSource == null) {
			throw new IllegalArgumentException("Please specify a source document/directory!");
		}
		// Parsing of command line ended successfully!
	}

	/**
	 * Extract the next argument from the array, while checking to see that the
	 * array size is not exceeded. Throw a friendly error message in case the arg
	 * is missing.
	 *
	 * @param i
	 *          Argument index.
	 * @param args
	 *          Array of command line arguments.
	 *
	 * @return The argument with the specified index.
	 *
	 * @throws IllegalArgumentException
	 *           If an argument is invalid.
	 */
	private String getArg(int i, String args[]) throws IllegalArgumentException {

		if (i < args.length) {
			return args[i];
		} else
			throw new IllegalArgumentException("I'm sorry, the commandline ended abnormally");
	}

}
